#include "POperators.h"
#include "Helpers.h"

using namespace std;
using namespace AMDiS;

Phase_SOT::Phase_SOT(DOFVectorBase<double>* phaseDV_,double fac_) :
	SecondOrderTerm(6), phaseDV(phaseDV_), fac(fac_)
{
	TEST_EXIT(phaseDV_)("no vector phase!\n");
	auxFeSpaces.insert(phaseDV_->getFeSpace());
};
void Phase_SOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) {
	getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void Phase_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad) {
	getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};
void Phase_SOT::getLALt(const ElInfo *elInfo,
			vector<mtl::dense2D<double> > &LALt) const
{
	const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());
    
	for (int iq = 0; iq < nPoints; iq++)
		l1lt(Lambda, LALt[iq], f(iq) * phase[iq] * fac);
};
void Phase_SOT::eval(int nPoints,
			const mtl::dense_vector<double>& uhAtQP,
			const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			mtl::dense_vector<double>& result,
			double opFactor)
{	
	if (num_rows(D2UhAtQP) > 0) {
		for (int iq = 0; iq < nPoints; iq++) {
			double feval = f(iq) * phase[iq] * opFactor * fac;
			double resultQP = 0.0;
			for (int i = 0; i < dimOfWorld; i++)
				resultQP += D2UhAtQP[iq][i][i];	
			result[iq] += feval * resultQP;
		}
	}
};
void Phase_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			std::vector<WorldVector<double> > &result)
{
	int nPoints = grdUhAtQP.size();
	for (int iq = 0; iq < nPoints; iq++) {
		double factor = f(iq) * phase[iq] * fac;
		axpy(factor, grdUhAtQP[iq], result[iq]);
	}
};
double Phase_SOT::f(const int iq) const
{
	return 1.0;
};

/* ----------------------------------------------------------- */

Phase_FOT::Phase_FOT(DOFVectorBase<double>* phaseDV_,double fac_) :
  FirstOrderTerm(6), phaseDV(phaseDV_), fac(fac_)
{
  TEST_EXIT(phaseDV_)("no vector phase!\n");
  auxFeSpaces.insert(phaseDV_->getFeSpace());
};
void Phase_FOT::initElement(const ElInfo* elInfo, 
      SubAssembler* subAssembler,
      Quadrature *quad) {
  getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void Phase_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
      SubAssembler* subAssembler,
      Quadrature *quad) {
  getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};
void Phase_FOT::getLb(const ElInfo *elInfo, 
            vector<mtl::dense_vector<double> >& result) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
    
    for (int iq = 0; iq < nPoints; iq++) {
        lb(Lambda, f(iq), result[iq], phase[iq]*fac);
    }
};
void Phase_FOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    double factor = fac * opFactor;
    for (int iq = 0; iq < nPoints; iq++) {
        double resultQP = grdUhAtQP[iq] * f(iq);
        result[iq] += phase[iq] * factor * resultQP;
    }
};
WorldVector<double> Phase_FOT::f(const int iq) const
{
  WorldVector<double> result;
  result.set(1.0);
  return result;
};

/* ----------------------------------------------------------- */

Phase_ZOT::Phase_ZOT(DOFVectorBase<double>* phaseDV_, double fac_) : ZeroOrderTerm(6),
	phaseDV(phaseDV_), fac(fac_)
{	
	TEST_EXIT(phaseDV_)("No vector Phase!\n");
	
	auxFeSpaces.insert(phaseDV_->getFeSpace());
};
void Phase_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void Phase_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};
void Phase_ZOT::getC(const ElInfo *elInfo, int nPoints, ElementVector &C)
{ 
	for (int iq = 0; iq < nPoints; iq++)
		C[iq] += f(iq) * phase[iq] * fac;
};
void Phase_ZOT::eval(int nPoints,
	const mtl::dense_vector<double>& uhAtQP,
	const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	mtl::dense_vector<double>& result,
	double opFactor)
{
	double factor = opFactor * fac;
	for (int iq = 0; iq < nPoints; iq++)
		result[iq] += factor * f(iq) * phase[iq] * uhAtQP[iq];
};
double Phase_ZOT::f(const int iq) const
{
	return 1.0;
};

/* ----------------------------------------------------------- */

PhaseInverse_ZOT::PhaseInverse_ZOT(DOFVectorBase<double>* phaseDV_, double fac_) : ZeroOrderTerm(6),
	phaseDV(phaseDV_), fac(fac_), component(0)
{	
	TEST_EXIT(phaseDV_)("No vector Phase!\n");
	facVec = new WorldVector<double>();
	facVec->set(1.0);

	auxFeSpaces.insert(phaseDV_->getFeSpace());
};
PhaseInverse_ZOT::PhaseInverse_ZOT(DOFVectorBase<double>* phaseDV_, double fac_, WorldVector<double> *facVec_, int component_) : ZeroOrderTerm(6),
	phaseDV(phaseDV_), fac(fac_), component(component_), facVec(facVec_)
{	
	TEST_EXIT(phaseDV_)("No vector Phase!\n");
	
	auxFeSpaces.insert(phaseDV_->getFeSpace());
};
void PhaseInverse_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void PhaseInverse_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};
void PhaseInverse_ZOT::getC(const ElInfo *, int nPoints, ElementVector &C)
{ 
	double factor = fac * (*facVec)[component];
	for (int iq = 0; iq < nPoints; iq++)
		C[iq] += f(iq) * factor;
};
void PhaseInverse_ZOT::eval(int nPoints,
				const mtl::dense_vector<double>& uhAtQP,
				const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				mtl::dense_vector<double>& result,
				double opFactor) 
{
	double factor = opFactor * fac * (*facVec)[component];
	for (int iq = 0; iq < nPoints; iq++)
		result[iq] += factor * f(iq) * uhAtQP[iq];
};
double PhaseInverse_ZOT::f(const int iq) const
{
	return std::max(0.0, std::min( 1.0, (1.0-phase[iq]) ) );
};

/* ----------------------------------------------------------- */

Pow3_ZOT::Pow3_ZOT(DOFVectorBase<double> *rhoDV_, double fac_) :
	ZeroOrderTerm(3), rhoDV(rhoDV_), fac(fac_)
{
	TEST_EXIT(rhoDV_)("No vector rho!\n");
	auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void Pow3_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void Pow3_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
void Pow3_ZOT::getC(const ElInfo *elInfo, int nPoints, ElementVector &C)
{ 
	for (int iq = 0; iq < nPoints; iq++)
		C[iq] += f(iq) * fac;
};
void Pow3_ZOT::eval(int nPoints,
	const mtl::dense_vector<double>& uhAtQP,
	const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	mtl::dense_vector<double>& result,
	double opFactor)
{
	double factor = opFactor * fac;
	for (int iq = 0; iq < nPoints; iq++)
		result[iq] += factor * f(iq) * uhAtQP[iq];
};
double Pow3_ZOT::f(const int iq) const
{
	return rho[iq]*sqr(rho[iq]);
};

/* ----------------------------------------------------------- */

Pow3Phase_ZOT::Pow3Phase_ZOT(DOFVectorBase<double>* phaseDV, DOFVectorBase<double> *rhoDV_, double fac_) :
	Phase_ZOT(phaseDV, fac_), rhoDV(rhoDV_)
{
	TEST_EXIT(rhoDV_)("No vector rho!\n");
	auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void Pow3Phase_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	Phase_ZOT::initElement(elInfo,subAssembler,quad);
	getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void Pow3Phase_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	Phase_ZOT::initElement(largeElInfo,smallElInfo,subAssembler,quad);
	getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
double Pow3Phase_ZOT::f(const int iq) const
{
	return rho[iq]*sqr(rho[iq]);
};

/* ----------------------------------------------------------- */

Pow2_ZOT::Pow2_ZOT(DOFVectorBase<double> *rhoDV_, double fac_) :
	ZeroOrderTerm(2), rhoDV(rhoDV_), fac(fac_)
{
	TEST_EXIT(rhoDV_)("No vector rho!\n");
	auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void Pow2_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void Pow2_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
void Pow2_ZOT::getC(const ElInfo *elInfo, int nPoints, ElementVector &C)
{ 
	for (int iq = 0; iq < nPoints; iq++)
		C[iq] += f(iq) * fac;
};
void Pow2_ZOT::eval(int nPoints,
	const mtl::dense_vector<double>& uhAtQP,
	const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	mtl::dense_vector<double>& result,
	double opFactor)
{
	double factor = opFactor * fac;
	for (int iq = 0; iq < nPoints; iq++)
		result[iq] += factor * f(iq) * uhAtQP[iq];
};
double Pow2_ZOT::f(const int iq) const
{
	return sqr(rho[iq]);
};

/* ----------------------------------------------------------- */

Pow2Phase_ZOT::Pow2Phase_ZOT(DOFVectorBase<double>* phaseDV, DOFVectorBase<double> *rhoDV_, double fac_) :
	Phase_ZOT(phaseDV, fac_), rhoDV(rhoDV_)
{
	TEST_EXIT(rhoDV_)("No vector rho!\n");
	auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void Pow2Phase_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	Phase_ZOT::initElement(elInfo,subAssembler,quad);
	getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void Pow2Phase_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	Phase_ZOT::initElement(largeElInfo,smallElInfo,subAssembler,quad);
	getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
double Pow2Phase_ZOT::f(const int iq) const
{
	return sqr(rho[iq]);
};
/* ----------------------------------------------------------- 
   1/(rho+0.9)
*/

ConstrainedFrac_ZOT::ConstrainedFrac_ZOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *rhoDV_, double fac_) :
	Phase_ZOT(phaseDV_,fac_), rhoDV(rhoDV_)
{
	TEST_EXIT(rhoDV_)("No vector rho!\n");
	auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void ConstrainedFrac_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	Phase_ZOT::initElement(elInfo, subAssembler,quad);
	getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void ConstrainedFrac_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	Phase_ZOT::initElement(largeElInfo, smallElInfo, subAssembler,quad);
	getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
double ConstrainedFrac_ZOT::f(const int iq) const
{
	return 1.0/std::max(1.e-6, rho[iq]+0.9);
};

/* ----------------------------------------------------------- 
   -1/(rho+0.9)^2
*/

ConstrainedFracSqr_ZOT::ConstrainedFracSqr_ZOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *rhoDV_, double factor_) :
	Phase_ZOT(phaseDV_,factor_), rhoDV(rhoDV_)
{
	TEST_EXIT(rhoDV_)("No vector rho!\n");
	auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void ConstrainedFracSqr_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	Phase_ZOT::initElement(elInfo, subAssembler,quad);
	getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void ConstrainedFracSqr_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	Phase_ZOT::initElement(largeElInfo, smallElInfo, subAssembler,quad);
	getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
double ConstrainedFracSqr_ZOT::f(const int iq) const
{
	return -1.0/std::max(1.e-6, sqr(rho[iq]+0.9));
};


ConstrainedPowImpl_ZOT::ConstrainedPowImpl_ZOT(DOFVectorBase<double> *rhoDV_, double p_, double rhoMin_, double fac_) :
    ZeroOrderTerm(2), rhoDV(rhoDV_), p(p_), rhoMin(rhoMin_), fac(fac_)
{
    TEST_EXIT(rhoDV_)("No vector rho!\n");
    auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void ConstrainedPowImpl_ZOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) 
{
    getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void ConstrainedPowImpl_ZOT::initElement(const ElInfo* largeElInfo,
            const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad)
{
    getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
void ConstrainedPowImpl_ZOT::getC(const ElInfo *elInfo, int nPoints, ElementVector &C)
{ 
    for (int iq = 0; iq < nPoints; iq++)
        C[iq] += f(iq)*fac;
};
void ConstrainedPowImpl_ZOT::eval(int nPoints,
    const mtl::dense_vector<double>& uhAtQP,
    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
    mtl::dense_vector<double>& result,
    double opFactor)
{
    double factor = opFactor*fac;
    for (int iq = 0; iq < nPoints; iq++)
        result[iq] += factor * f(iq) * uhAtQP[iq];
};
double ConstrainedPowImpl_ZOT::f(const int iq) const
{
    return 2.0*p*Helpers::toInt(0.0<=rhoMin-rho[iq]);
};


ConstrainedPowExpl_ZOT::ConstrainedPowExpl_ZOT(DOFVectorBase<double> *rhoDV_, double p_, double rhoMin_, double fac_) :
    ZeroOrderTerm(2), rhoDV(rhoDV_), p(p_), rhoMin(rhoMin_), fac(fac_)
{
    TEST_EXIT(rhoDV_)("No vector rho!\n");
    auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void ConstrainedPowExpl_ZOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) 
{
    getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void ConstrainedPowExpl_ZOT::initElement(const ElInfo* largeElInfo,
            const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad)
{
    getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
void ConstrainedPowExpl_ZOT::getC(const ElInfo *elInfo, int nPoints, ElementVector &C)
{ 
    for (int iq = 0; iq < nPoints; iq++)
        C[iq] += f(iq)*fac;
};
void ConstrainedPowExpl_ZOT::eval(int nPoints,
    const mtl::dense_vector<double>& uhAtQP,
    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
    mtl::dense_vector<double>& result,
    double opFactor)
{
    double factor = opFactor*fac;
    for (int iq = 0; iq < nPoints; iq++)
        result[iq] += factor * f(iq) * uhAtQP[iq];
};
double ConstrainedPowExpl_ZOT::f(const int iq) const
{
    return -2.0*p*std::max(0.0, rhoMin-rho[iq]);
};

ConstrainedPow_ZOT::ConstrainedPow_ZOT(DOFVectorBase<double> *rhoDV_, double p_, double rhoMin_, double fac_) :
    ZeroOrderTerm(2), rhoDV(rhoDV_), p(p_), rhoMin(rhoMin_), fac(fac_)
{
    TEST_EXIT(rhoDV_)("No vector rho!\n");
    auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void ConstrainedPow_ZOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) 
{
    getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void ConstrainedPow_ZOT::initElement(const ElInfo* largeElInfo,
            const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad)
{
    getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
void ConstrainedPow_ZOT::getC(const ElInfo *elInfo, int nPoints, ElementVector &C)
{ 
    for (int iq = 0; iq < nPoints; iq++)
        C[iq] += f(iq)*fac;
};
void ConstrainedPow_ZOT::eval(int nPoints,
    const mtl::dense_vector<double>& uhAtQP,
    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
    mtl::dense_vector<double>& result,
    double opFactor)
{
    double factor = opFactor*fac;
    for (int iq = 0; iq < nPoints; iq++)
        result[iq] += factor * f(iq) * uhAtQP[iq];
};
double ConstrainedPow_ZOT::f(const int iq) const
{
    return -2.0*p*std::max(0.0, rhoMin-rho[iq]);
};

/* -----------------------------------------------------------
   phi*rho
   --------- */

FactorPhase_ZOT::FactorPhase_ZOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *rhoDV_, double fac_) :
	Phase_ZOT(phaseDV_,fac_), rhoDV(rhoDV_)
{
	TEST_EXIT(rhoDV_)("No vector rho!\n");
	auxFeSpaces.insert(rhoDV_->getFeSpace());
};

void FactorPhase_ZOT::initElement(const ElInfo* elInfo, 
			SubAssembler* subAssembler,
			Quadrature *quad) 
{
	Phase_ZOT::initElement(elInfo, subAssembler,quad);
	getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void FactorPhase_ZOT::initElement(const ElInfo* largeElInfo,
			const ElInfo* smallElInfo,
			SubAssembler* subAssembler,
			Quadrature *quad)
{
	Phase_ZOT::initElement(largeElInfo, smallElInfo, subAssembler,quad);
	getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
double FactorPhase_ZOT::f(const int iq) const
{
	return rho[iq];
};

/* ----------------------------------------------------------- */

WorldVecAndVecFct_FOT::WorldVecAndVecFct_FOT(WorldVector<DOFVectorBase<double>*> vecs_, DOFVectorBase<double> *phaseDV_, double fac_)
	: FirstOrderTerm(2), phaseDV(phaseDV_), fac(fac_)
{
	numVecs=vecs_.size();
	TEST_EXIT(numVecs==2 || numVecs==3)("Only Dim=2 or Dim=3 possible\n");
	
	TEST_EXIT(phaseDV_)("phaseDV is NULL!\n");
	for (int i = 0; i < numVecs; i++) {
		TEST_EXIT(vecs_[i])("One vector is NULL!\n");
	
		auxFeSpaces.insert(vecs_[i]->getFeSpace());
	}
	
	vec0DV=vecs_[0];
	vec1DV=vecs_[1];
	if(numVecs>=3) vec2DV=vecs_[2];
};
void WorldVecAndVecFct_FOT::initElement(const ElInfo* elInfo, 
					SubAssembler* subAssembler,
					Quadrature *quad) 
{
	getVectorAtQPs(vec0DV, elInfo, subAssembler, quad, vec0);
	getVectorAtQPs(vec1DV, elInfo, subAssembler, quad, vec1);
	if(numVecs>=3) getVectorAtQPs(vec2DV, elInfo, subAssembler, quad, vec2);
	getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void WorldVecAndVecFct_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
					SubAssembler* subAssembler,
					Quadrature *quad) 
{
	getVectorAtQPs(vec0DV, smallElInfo, largeElInfo, subAssembler, quad, vec0);
	getVectorAtQPs(vec1DV, smallElInfo, largeElInfo, subAssembler, quad, vec1);
	if(numVecs>=3) getVectorAtQPs(vec2DV, smallElInfo, largeElInfo, subAssembler, quad, vec2);
	getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};
void WorldVecAndVecFct_FOT::getLb(const ElInfo *elInfo, 
			vector<mtl::dense_vector<double> >& result) const
{
	const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
	
	for (int iq = 0; iq < nPoints; iq++) {		
		WorldVector<double> vec;
		vec[0]=vec0[iq];
		vec[1]=vec1[iq];
		if(numVecs>=3) vec[2]=vec2[iq];
		lb(Lambda, vec, result[iq], phase[iq]*fac);
	}
};
void WorldVecAndVecFct_FOT::eval(int nPoints,
			const mtl::dense_vector<double>& uhAtQP,
			const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			mtl::dense_vector<double>& result,
			double opFactor)
{	
	double factor = fac * opFactor;
	for (int iq = 0; iq < nPoints; iq++) {
		double resultQP = 0.0;
				
		resultQP += grdUhAtQP[iq][0] * vec0[iq];
		resultQP += grdUhAtQP[iq][1] * vec1[iq];
		if(numVecs>=3) resultQP += grdUhAtQP[iq][2] * vec2[iq];
		
		result[iq] += phase[iq] * factor * resultQP;
	}
};

/* ----------------------------------------------------------- */

WorldVec_FOT::WorldVec_FOT(WorldVector<DOFVector<double>*> vecs_, double fac_)
	: FirstOrderTerm(2), fac(fac_)
{
	numVecs=vecs_.size();
	TEST_EXIT(numVecs==2 || numVecs==3)("Only Dim=2 or Dim=3 possible\n");
	
	for (int i = 0; i < numVecs; i++) {
		TEST_EXIT(vecs_[i])("One vector is NULL!\n");	
		auxFeSpaces.insert(vecs_[i]->getFeSpace());
	}
	
	vec0DV=vecs_[0];
	vec1DV=vecs_[1];
	if(numVecs>=3) vec2DV=vecs_[2];
};
void WorldVec_FOT::initElement(const ElInfo* elInfo, 
					SubAssembler* subAssembler,
					Quadrature *quad) 
{
	getVectorAtQPs(vec0DV, elInfo, subAssembler, quad, vec0);
	getVectorAtQPs(vec1DV, elInfo, subAssembler, quad, vec1);
	if(numVecs>=3) getVectorAtQPs(vec2DV, elInfo, subAssembler, quad, vec2);
};
void WorldVec_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
					SubAssembler* subAssembler,
					Quadrature *quad) 
{
	getVectorAtQPs(vec0DV, smallElInfo, largeElInfo, subAssembler, quad, vec0);
	getVectorAtQPs(vec1DV, smallElInfo, largeElInfo, subAssembler, quad, vec1);
	if(numVecs>=3) getVectorAtQPs(vec2DV, smallElInfo, largeElInfo, subAssembler, quad, vec2);
};
void WorldVec_FOT::getLb(const ElInfo *elInfo, 
			vector<mtl::dense_vector<double> >& result) const
{
	const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
	
	for (int iq = 0; iq < nPoints; iq++) {
		WorldVector<double> vec;
		vec[0]=vec0[iq];
		vec[1]=vec1[iq];
		if(numVecs>=3) vec[2]=vec2[iq];
		lb(Lambda, vec, result[iq], fac);
	}
};
void WorldVec_FOT::eval(int nPoints,
			const mtl::dense_vector<double>& uhAtQP,
			const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			mtl::dense_vector<double>& result,
			double opFactor)
{	
	double factor = fac * opFactor;
	for (int iq = 0; iq < nPoints; iq++) {
		double resultQP = 0.0;
				
		resultQP += grdUhAtQP[iq][0] * vec0[iq];
		resultQP += grdUhAtQP[iq][1] * vec1[iq];
		if(numVecs>=3) resultQP += grdUhAtQP[iq][2] * vec2[iq];
		
		result[iq] += factor * resultQP;
	}
};

/* ----------------------------------------------------------- */

WorldVector_FOT::WorldVector_FOT(DOFVector<WorldVector<double> >* dv, double fac_)
: FirstOrderTerm(1),
  vec(dv),
  fac(fac_)
{
  TEST_EXIT(dv)("No vector!\n");

  auxFeSpaces.insert(dv->getFeSpace());
}


void WorldVector_FOT::initElement(const ElInfo* elInfo,
				  SubAssembler* subAssembler,
				  Quadrature *quad)
{
  getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
}


void WorldVector_FOT::initElement(const ElInfo* largeElInfo,
				  const ElInfo* smallElInfo,
				  SubAssembler* subAssembler,
				  Quadrature *quad)
{
  getVectorAtQPs(vec, smallElInfo, largeElInfo, subAssembler, quad, vecAtQPs);
}


void WorldVector_FOT::getLb(const ElInfo *elInfo,
			    vector<mtl::dense_vector<double> >& Lb) const
{
  const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
  const int nPoints = static_cast<int>(Lb.size());

  for (int iq = 0; iq < nPoints; iq++)
    lb(grdLambda, vecAtQPs[iq], Lb[iq], fac);
}


void WorldVector_FOT::eval(int nPoints,
			   const mtl::dense_vector<double>&,
			   const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			   const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			   mtl::dense_vector<double>& result,
		 	   double opFactor)
{
  if (num_rows(grdUhAtQP) > 0) {
    double fac_ = fac * opFactor;
    for (int iq = 0; iq < nPoints; iq++)
      result[iq] += (vecAtQPs[iq] * grdUhAtQP[iq]) * fac_;
  }
}

/* ----------------------------------------------------------- */

WorldVecPhase_FOT::WorldVecPhase_FOT(DOFVectorBase<double> *phaseDV_, WorldVector<DOFVector<double>*> vecs_,double fac_)
    : FirstOrderTerm(2), phaseDV(phaseDV_), fac(fac_)
{
    numVecs=vecs_.size();
    TEST_EXIT(numVecs==2 || numVecs==3)("Only Dim=2 or Dim=3 possible\n");
    
    TEST_EXIT(phaseDV_)("phaseDV is NULL!\n");
    for (int i = 0; i < numVecs; i++) {
        TEST_EXIT(vecs_[i])("One vector is NULL!\n");
    
        auxFeSpaces.insert(vecs_[i]->getFeSpace());
    }
    auxFeSpaces.insert(phaseDV_->getFeSpace());
    
    vec0DV=vecs_[0];
    vec1DV=vecs_[1];
    if(numVecs>=3) vec2DV=vecs_[2];
};
void WorldVecPhase_FOT::initElement(const ElInfo* elInfo, 
                    SubAssembler* subAssembler,
                    Quadrature *quad) 
{
    getVectorAtQPs(vec0DV, elInfo, subAssembler, quad, vec0);
    getVectorAtQPs(vec1DV, elInfo, subAssembler, quad, vec1);
    if(numVecs>=3) getVectorAtQPs(vec2DV, elInfo, subAssembler, quad, vec2);
    getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void WorldVecPhase_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
                    SubAssembler* subAssembler,
                    Quadrature *quad) 
{
    getVectorAtQPs(vec0DV, smallElInfo, largeElInfo, subAssembler, quad, vec0);
    getVectorAtQPs(vec1DV, smallElInfo, largeElInfo, subAssembler, quad, vec1);
    if(numVecs>=3) getVectorAtQPs(vec2DV, smallElInfo, largeElInfo, subAssembler, quad, vec2);
    getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};
void WorldVecPhase_FOT::getLb(const ElInfo *elInfo, 
            vector<mtl::dense_vector<double> >& result) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
    
    for (int iq = 0; iq < nPoints; iq++) {      
        WorldVector<double> vec;
        vec[0]=vec0[iq];
        vec[1]=vec1[iq];
        if(numVecs>=3) vec[2]=vec2[iq];
        lb(Lambda, vec, result[iq], phase[iq]*fac);
    }
};
void WorldVecPhase_FOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    double factor = fac * opFactor;
    for (int iq = 0; iq < nPoints; iq++) {
        double resultQP = 0.0;
                
        resultQP += grdUhAtQP[iq][0] * vec0[iq];
        resultQP += grdUhAtQP[iq][1] * vec1[iq];
        if(numVecs>=3) resultQP += grdUhAtQP[iq][2] * vec2[iq];
        
        result[iq] += phase[iq] * factor * resultQP;
    }
};

/* ----------------------------------------------------------- */

// vec1*vec2*d/dxi(...)
VecAndPartialDerivative_FOT::VecAndPartialDerivative_FOT(DOFVectorBase<double> *dv1,
      int component_, double fac_)
: FirstOrderTerm(3), vec1DV(dv1), fac(fac_), component(component_)
{
  auxFeSpaces.insert(vec1DV->getFeSpace());

  setB(component);
}

void VecAndPartialDerivative_FOT::initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad) 
{ 
  getVectorAtQPs(vec1DV, elInfo, subAssembler, quad, vec1);
};
void VecAndPartialDerivative_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
  SubAssembler* subAssembler,
  Quadrature *quad)
{ 
  getVectorAtQPs(vec1DV, smallElInfo, largeElInfo, subAssembler, quad, vec1);
};
void VecAndPartialDerivative_FOT::getLb(const ElInfo *elInfo,
      vector<mtl::dense_vector<double> >& result) const
{
  const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
    
  for(int iq = 0; iq < nPoints; iq++) {
    lb_one(Lambda, result[iq], vec1[iq] * fac);
  }
};
void VecAndPartialDerivative_FOT::eval(int nPoints,
      const mtl::dense_vector<double>& uhAtQP,
      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
      mtl::dense_vector<double>& result,
      double opFactor)
{
  for(int iq = 0; iq < nPoints; iq++) 
    result[iq] += opFactor * vec1[iq] * grdUhAtQP[iq][component] * fac;
};

/* ----------------------------------------------------------- */

// vec1*vec2*d/dxi(...)
PhaseAndViscosityPartialDerivativeBoundary_FOT::PhaseAndViscosityPartialDerivativeBoundary_FOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *chDV_,
      int component_, int normalComponent_, double viscosity1_, double viscosity2_)
: FirstOrderTerm(3), phaseDV(phaseDV_), chDV(chDV_), component(component_), normalComponent(normalComponent_), v1(viscosity1_), v2(viscosity2_)
{
  auxFeSpaces.insert(phaseDV->getFeSpace());
  auxFeSpaces.insert(chDV->getFeSpace());

  setB(component);
}

void PhaseAndViscosityPartialDerivativeBoundary_FOT::initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad) 
{ 
  getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
  getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);

  WorldVector<double> normal;
  int side= -1;
  for (int i= 0; i < elInfo->getMesh()->getDim()+1; ++i) {
    if(elInfo->getBoundary(i) > 2) {
      side= i;
      break;
    }
  }
//   elInfo->getNormal(side, normal);
  normalJ = 1.0; //normal[normalComponent];
};
void PhaseAndViscosityPartialDerivativeBoundary_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
  SubAssembler* subAssembler,
  Quadrature *quad)
{ 
  getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
  getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);

  WorldVector<double> normal;
  int side= -1;
  for (int i= 0; i < largeElInfo->getMesh()->getDim()+1; ++i) {
    if(largeElInfo->getBoundary(i) > 2) {
      side= i;
      break;
    }
  }
//   largeElInfo->getNormal(side, normal);
  normalJ = 1.0; //normal[normalComponent];
};
void PhaseAndViscosityPartialDerivativeBoundary_FOT::getLb(const ElInfo *elInfo,
      vector<mtl::dense_vector<double> >& result) const
{
  const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
  const int nPoints = static_cast<int>(result.size());
    
  for(int iq = 0; iq < nPoints; iq++) {
    lb_one(Lambda, result[iq], phase[iq] * normalJ * f(iq));
  }
};
void PhaseAndViscosityPartialDerivativeBoundary_FOT::eval(int nPoints,
      const mtl::dense_vector<double>& uhAtQP,
      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
      mtl::dense_vector<double>& result,
      double opFactor)
{
  for(int iq = 0; iq < nPoints; iq++) 
    result[iq] += opFactor * phase[iq] * grdUhAtQP[iq][component] * normalJ * f(iq);
};

double PhaseAndViscosityPartialDerivativeBoundary_FOT::f(const int iq) const
{
    double ph= std::max(-1.0,std::min(1.0,ch[iq]));
    double viscosity= v1*0.5*(ph+1.0)-v2*0.5*(ph-1.0);
    return viscosity;
};

/* ----------------------------------------------------------- */

// vec1*vec2*d/dxi(...)
ViscosityPartialDerivativeBoundary_FOT::ViscosityPartialDerivativeBoundary_FOT(DOFVectorBase<double> *chDV_,
      int component_, int normalComponent_, double viscosity1_, double viscosity2_)
: FirstOrderTerm(3), chDV(chDV_), component(component_), normalComponent(normalComponent_), v1(viscosity1_), v2(viscosity2_)
{
  auxFeSpaces.insert(chDV->getFeSpace());

  setB(component);
}

void ViscosityPartialDerivativeBoundary_FOT::initElement(const ElInfo* elInfo, 
        SubAssembler* subAssembler,
        Quadrature *quad) 
{ 
  getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);

  WorldVector<double> normal;
  int side= -1;
  for (int i= 0; i < elInfo->getMesh()->getDim()+1; ++i) {
    if(elInfo->getBoundary(i) > 2) {
      side= i;
      break;
    }
  }
//   elInfo->getNormal(side, normal);
  normalJ = 1.0; //normal[normalComponent];
};
void ViscosityPartialDerivativeBoundary_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
  SubAssembler* subAssembler,
  Quadrature *quad)
{ 
  getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);

  WorldVector<double> normal;
  int side= -1;
  for (int i= 0; i < largeElInfo->getMesh()->getDim()+1; ++i) {
    if(largeElInfo->getBoundary(i) > 2) {
      side= i;
      break;
    }
  }
//   largeElInfo->getNormal(side, normal);
  normalJ = 1.0; //normal[normalComponent];
};
void ViscosityPartialDerivativeBoundary_FOT::getLb(const ElInfo *elInfo,
      vector<mtl::dense_vector<double> >& result) const
{
  const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
  const int nPoints = static_cast<int>(result.size());
    
  for(int iq = 0; iq < nPoints; iq++) {
    lb_one(Lambda, result[iq], normalJ * f(iq));
  }
};
void ViscosityPartialDerivativeBoundary_FOT::eval(int nPoints,
      const mtl::dense_vector<double>& uhAtQP,
      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
      mtl::dense_vector<double>& result,
      double opFactor)
{
  for(int iq = 0; iq < nPoints; iq++) 
    result[iq] += opFactor * grdUhAtQP[iq][component] * normalJ * f(iq);
};

double ViscosityPartialDerivativeBoundary_FOT::f(const int iq) const
{
    double ph= std::max(-1.0,std::min(1.0,ch[iq]));
    double viscosity= v1*0.5*(ph+1.0)-v2*0.5*(ph-1.0);
    return viscosity;
};
/* ----------------------------------------------------------- */

// vec1*vec2*d/dxi(...)
Vec2ProductPartial_FOT::Vec2ProductPartial_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
			int component_, double fac_)
: FirstOrderTerm(3), vec1DV(dv1), vec2DV(dv2), fac(fac_), component(component_)
{
	auxFeSpaces.insert(vec1DV->getFeSpace());
	auxFeSpaces.insert(vec2DV->getFeSpace());

	setB(component);
}

void Vec2ProductPartial_FOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature *quad) 
{ 
	getVectorAtQPs(vec1DV, elInfo, subAssembler, quad, vec1);
	getVectorAtQPs(vec2DV, elInfo, subAssembler, quad, vec2);
};
void Vec2ProductPartial_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler* subAssembler,
	Quadrature *quad)
{ 
	getVectorAtQPs(vec1DV, smallElInfo, largeElInfo, subAssembler, quad, vec1);
	getVectorAtQPs(vec2DV, smallElInfo, largeElInfo, subAssembler, quad, vec2);
};
void Vec2ProductPartial_FOT::getLb(const ElInfo *elInfo,
			vector<mtl::dense_vector<double> >& result) const
{
	const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
    
	for(int iq = 0; iq < nPoints; iq++) {
		lb_one(Lambda, result[iq], vec1[iq] * vec2[iq] * fac);
	}
};
void Vec2ProductPartial_FOT::eval(int nPoints,
			const mtl::dense_vector<double>& uhAtQP,
			const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			mtl::dense_vector<double>& result,
			double opFactor)
{
	for(int iq = 0; iq < nPoints; iq++) 
		result[iq] += opFactor * vec1[iq] * vec2[iq] * grdUhAtQP[iq][component] * fac;
};

/* ----------------------------------------------------------- */

PartialDerivative_FOT::PartialDerivative_FOT(int component_, double fac_)
	: FirstOrderTerm(1), component(component_), fac(fac_)
{
	setB(component);
};
void PartialDerivative_FOT::initElement(const ElInfo* elInfo, 
					SubAssembler* subAssembler,
					Quadrature *quad) {};
void PartialDerivative_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
					SubAssembler* subAssembler,
					Quadrature *quad) {};
void PartialDerivative_FOT::getLb(const ElInfo *elInfo, 
			vector<mtl::dense_vector<double> >& result) const
{
	const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
	
	for (int iq = 0; iq < nPoints; iq++) {	
		lb_one(Lambda, result[iq], fac);
	}
};
void PartialDerivative_FOT::eval(int nPoints,
			const mtl::dense_vector<double>& uhAtQP,
			const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			mtl::dense_vector<double>& result,
			double opFactor)
{	
	for (int iq = 0; iq < nPoints; iq++) {		
		result[iq] += opFactor * grdUhAtQP[iq][component] * fac;
	}
};

/* ----------------------------------------------------------- */

PartialDerivative_ZOT::PartialDerivative_ZOT(DOFVectorBase<double> *vecDV_, int component_, double fac_)
	: ZeroOrderTerm(1), vecDV(vecDV_), fac(fac_), component(component_)
{
	TEST_EXIT(vecDV_)("No first vector!\n");
	auxFeSpaces.insert(vecDV_->getFeSpace());

	facVec = new WorldVector<double>(); facVec->set(1.0);
	component2 = 0;
};
PartialDerivative_ZOT::PartialDerivative_ZOT(DOFVectorBase<double> *vecDV_, int component_, double fac_, WorldVector<double> *facVec_, int component2_)
	: ZeroOrderTerm(1), vecDV(vecDV_), fac(fac_), facVec(facVec_), component(component_), component2(component2_)
{
	TEST_EXIT(vecDV_)("No first vector!\n");
	auxFeSpaces.insert(vecDV_->getFeSpace());
};
void PartialDerivative_ZOT::initElement(const ElInfo* elInfo, 
					SubAssembler* subAssembler,
					Quadrature *quad)
{
  getGradientsAtQPs(vecDV, elInfo, subAssembler, quad, grad);
};
void PartialDerivative_ZOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
					SubAssembler* subAssembler,
					Quadrature *quad)
{
  getGradientsAtQPs(vecDV, smallElInfo, largeElInfo, subAssembler, quad, grad);
};
void PartialDerivative_ZOT::getC(const ElInfo *, int nPoints, ElementVector &C)
{ 	
	double factor = fac * (*facVec)[component2];
	for (int iq = 0; iq < nPoints; iq++)
		C[iq] += grad[iq][component]*factor;
};
void PartialDerivative_ZOT::eval(int nPoints,
				const mtl::dense_vector<double>& uhAtQP,
				const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				mtl::dense_vector<double>& result,
				double opFactor)
{	
	double factor = opFactor * fac * (*facVec)[component2];
	for (int iq = 0; iq < nPoints; iq++) {		
		result[iq] += grad[iq][component] * uhAtQP[iq] * factor;
	}
};


/* ----------------------------------------------------------- */

VecAndPartialDerivative_ZOT::VecAndPartialDerivative_ZOT(DOFVectorBase<double> *vecDV_, DOFVectorBase<double> *gradDV_, int component_,  double fac_)
: ZeroOrderTerm(2),
  vecDV(vecDV_),
  gradDV(gradDV_),
  component(component_),
  fac(fac_)
{
  TEST_EXIT(vecDV_)("No value vector!\n");
  TEST_EXIT(gradDV_)("No gradient vector!\n");

  auxFeSpaces.insert(vecDV_->getFeSpace());
  auxFeSpaces.insert(gradDV_->getFeSpace());
}


void VecAndPartialDerivative_ZOT::initElement(const ElInfo* elInfo,
					      SubAssembler* subAssembler,
					      Quadrature *quad)
{
  getVectorAtQPs(vecDV, elInfo, subAssembler, quad, vec);
  getGradientsAtQPs(gradDV, elInfo, subAssembler, quad, grad);
}


void VecAndPartialDerivative_ZOT::initElement(const ElInfo* largeElInfo,
					      const ElInfo* smallElInfo,
					      SubAssembler* subAssembler,
					      Quadrature *quad)
{
  getVectorAtQPs(vecDV, smallElInfo, largeElInfo, subAssembler, quad, vec);
  getGradientsAtQPs(gradDV, smallElInfo, largeElInfo, subAssembler, quad, grad);
}


void VecAndPartialDerivative_ZOT::getC(const ElInfo *, int nPoints, ElementVector &C)
{
  for (int iq = 0; iq < nPoints; iq++)
    C[iq] += vec[iq] * grad[iq][component] * fac;
}


void VecAndPartialDerivative_ZOT::eval(int nPoints,
				       const mtl::dense_vector<double>& uhAtQP,
				       const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				       const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				       mtl::dense_vector<double>& result,
				       double opFactor)
{
  double fac_ = fac * opFactor;
  for (int iq = 0; iq < nPoints; iq++)
    result[iq] += vec[iq] * grad[iq][component] * uhAtQP[iq] * fac_;
}

/* ----------------------------------------------------------- */

Vec2AndPartialDerivative_ZOT::Vec2AndPartialDerivative_ZOT(DOFVectorBase<double> *vec1DV_, DOFVectorBase<double> *vec2DV_, DOFVectorBase<double> *gradDV_,int component_,  double fac_)
	: ZeroOrderTerm(3), vec1DV(vec1DV_), vec2DV(vec2DV_), gradDV(gradDV_), fac(fac_), component(component_)
{
	TEST_EXIT(vec1DV_)("No first vector!\n");
	TEST_EXIT(vec2DV_)("No second vector!\n");
	TEST_EXIT(gradDV_)("No gradient vector!\n");
	auxFeSpaces.insert(vec1DV_->getFeSpace());
	auxFeSpaces.insert(vec2DV_->getFeSpace());
	auxFeSpaces.insert(gradDV_->getFeSpace());
};
void Vec2AndPartialDerivative_ZOT::initElement(const ElInfo* elInfo,
					SubAssembler* subAssembler,
					Quadrature *quad)
{
	getVectorAtQPs(vec1DV, elInfo, subAssembler, quad, vec1);
	getVectorAtQPs(vec2DV, elInfo, subAssembler, quad, vec2);
	getGradientsAtQPs(gradDV, elInfo, subAssembler, quad, grad);
};
void Vec2AndPartialDerivative_ZOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
					SubAssembler* subAssembler,
					Quadrature *quad)
{
	getVectorAtQPs(vec1DV, smallElInfo, largeElInfo, subAssembler, quad, vec1);
	getVectorAtQPs(vec2DV, smallElInfo, largeElInfo, subAssembler, quad, vec2);
	getGradientsAtQPs(gradDV, smallElInfo, largeElInfo, subAssembler, quad, grad);
};
void Vec2AndPartialDerivative_ZOT::getC(const ElInfo *, int nPoints, ElementVector &C)
{
	for (int iq = 0; iq < nPoints; iq++)
		C[iq] += vec1[iq]*vec2[iq]*grad[iq][component]*fac;
};
void Vec2AndPartialDerivative_ZOT::eval(int nPoints,
				const mtl::dense_vector<double>& uhAtQP,
				const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				mtl::dense_vector<double>& result,
				double opFactor)
{
	for (int iq = 0; iq < nPoints; iq++) {
		result[iq] += (vec1[iq] * vec2[iq]) * grad[iq][component] * uhAtQP[iq] * opFactor * fac;
	}
};

/* ----------------------------------------------------------- */

Mobility_SOT::Mobility_SOT(DOFVectorBase<double> *rhoDV_, double factor_)
    : SecondOrderTerm(6), rhoDV(rhoDV_), factor(factor_)
{
    setSymmetric(true);
    
    TEST_EXIT(rhoDV)("No vector!\n");
    auxFeSpaces.insert(rhoDV->getFeSpace());
    eps= 1.e-5;
    
    rho0= 1.0;
    density= -0.3;
    Parameters::get("pfc->rho0",rho0, 1);
    Parameters::get("pfc->density",density, 1);
};
void Mobility_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(rhoDV, elInfo, subAssembler, quad, rho);
};
void Mobility_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(rhoDV, smallElInfo, largeElInfo, subAssembler, quad, rho);
};
void Mobility_SOT::getLALt(const ElInfo *elInfo,
            std::vector<mtl::dense2D<double> > &LALt) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());
    
    for (int iq = 0; iq < nPoints; iq++)
        l1lt(Lambda, LALt[iq], f(iq) * factor);
};
void Mobility_SOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            double feval = f(iq) * opFactor * factor;
            double resultQP = 0.0;
            for (int i = 0; i < dimOfWorld; i++)
                resultQP += D2UhAtQP[iq][i][i]; 
            result[iq] += feval * resultQP;
        }
    }
};
void Mobility_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
            std::vector<WorldVector<double> > &result)
{
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
        double fac = f(iq) * factor;
        axpy(fac, grdUhAtQP[iq], result[iq]);
    }
};
double Mobility_SOT::f(const int iq) const
{
    double mobility= 0.5*(3.0-rho[iq]/density)*rho0;
//     double phase = 0.5*(1.0-tanh(3.0*(eps-mobility)/0.1));
    return std::max(mobility, eps);
};

/* ----------------------------------------------------------- */

MobilityCH_SOT::MobilityCH_SOT(DOFVectorBase<double> *chDV_, double factor_)
    : SecondOrderTerm(4), chDV(chDV_), factor(factor_)
{
    setSymmetric(true);
    
    TEST_EXIT(chDV)("No vector!\n");
    auxFeSpaces.insert(chDV->getFeSpace());
    delta = 1.e-6;
};
void MobilityCH_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void MobilityCH_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
void MobilityCH_SOT::getLALt(const ElInfo *elInfo,
            std::vector<mtl::dense2D<double> > &LALt) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());
    
    for (int iq = 0; iq < nPoints; iq++)
        l1lt(Lambda, LALt[iq], f(iq) * factor);
};
void MobilityCH_SOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            double feval = f(iq) * opFactor * factor;
            double resultQP = 0.0;
            for (int i = 0; i < dimOfWorld; i++)
                resultQP += D2UhAtQP[iq][i][i]; 
            result[iq] += feval * resultQP;
        }
    }
};
void MobilityCH_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
            std::vector<WorldVector<double> > &result)
{
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
        double fac = f(iq) * factor;
        axpy(fac, grdUhAtQP[iq], result[iq]);
    }
};
double MobilityCH_SOT::f(const int iq) const
{
    double phase= std::max(-1.0,std::min(1.0,ch[iq]));
    double mobility= 0.25*sqr(sqr(phase)-1.0);
      return std::max(mobility, delta);
};


/* ----------------------------------------------------------- */

MobilityCH2_SOT::MobilityCH2_SOT(DOFVectorBase<double> *chDV_, double factor_)
    : SecondOrderTerm(4), chDV(chDV_), factor(factor_)
{
    setSymmetric(true);
    
    TEST_EXIT(chDV)("No vector!\n");
    auxFeSpaces.insert(chDV->getFeSpace());
    delta = 1.e-6;
};
void MobilityCH2_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void MobilityCH2_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
void MobilityCH2_SOT::getLALt(const ElInfo *elInfo,
            std::vector<mtl::dense2D<double> > &LALt) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());
    
    for (int iq = 0; iq < nPoints; iq++)
        l1lt(Lambda, LALt[iq], f(iq) * factor);
};
void MobilityCH2_SOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            double feval = f(iq) * opFactor * factor;
            double resultQP = 0.0;
            for (int i = 0; i < dimOfWorld; i++)
                resultQP += D2UhAtQP[iq][i][i]; 
            result[iq] += feval * resultQP;
        }
    }
};
void MobilityCH2_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
            std::vector<WorldVector<double> > &result)
{
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
        double fac = f(iq) * factor;
        axpy(fac, grdUhAtQP[iq], result[iq]);
    }
};
double MobilityCH2_SOT::f(const int iq) const
{
    double phase= std::max(0.0,std::min(1.0,ch[iq]));
  
    double mobility= 0.25*sqr(phase)*sqr(phase-1.0);
    return std::max(mobility, delta);
};

/* ----------------------------------------------------------- */

MobilityCHPhase_SOT::MobilityCHPhase_SOT(DOFVectorBase<double>* phaseDV_, DOFVectorBase<double> *chDV_, double factor_)
    : Phase_SOT(phaseDV_, factor_), chDV(chDV_)
{
    setSymmetric(true);
    
    TEST_EXIT(chDV)("No vector!\n");
    auxFeSpaces.insert(chDV->getFeSpace());
    delta = 1.e-6;
};
void MobilityCHPhase_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
	Phase_SOT::initElement(elInfo,subAssembler,quad);
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void MobilityCHPhase_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
	Phase_SOT::initElement(largeElInfo, smallElInfo,subAssembler,quad);
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
double MobilityCHPhase_SOT::f(const int iq) const
{
    double phase= std::max(-1.0,std::min(1.0,ch[iq]));
    double mobility= 0.25*sqr(sqr(phase)-1.0);
  
    return std::max(mobility, delta);
};


/* ----------------------------------------------------------- */

Mobility2CHPhase_FOT::Mobility2CHPhase_FOT(
  DOFVectorBase<double>* phaseDV_, 
  DOFVectorBase<double> *chDV_, 
  DOFVectorBase<double> *muDV_, 
  double factor_)
    : Phase_FOT(phaseDV_, factor_), chDV(chDV_), muDV(muDV_)
{  
  TEST_EXIT(chDV)("No vector!\n");
  auxFeSpaces.insert(chDV->getFeSpace());
  auxFeSpaces.insert(muDV->getFeSpace());
  delta = 1.e-6;
};
void Mobility2CHPhase_FOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
  Phase_FOT::initElement(elInfo,subAssembler,quad);
  getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
  getGradientsAtQPs(muDV, elInfo, subAssembler, quad, grdMu);
};
void Mobility2CHPhase_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
  Phase_FOT::initElement(largeElInfo, smallElInfo,subAssembler,quad);
  getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
  getGradientsAtQPs(muDV, smallElInfo, largeElInfo, subAssembler, quad, grdMu);
};
WorldVector<double> Mobility2CHPhase_FOT::f(const int iq) const
{
    double mobility= sqr(ch[iq])*ch[iq]-ch[iq];
    return mobility*grdMu[iq];
};

/* ----------------------------------------------------------- */


MatrixPhase_SOT::MatrixPhase_SOT(DOFVectorBase<double>* phaseDV_, int comp_, double fac_) :
    SecondOrderTerm(6), phaseDV(phaseDV_), comp(comp_), fac(fac_), symmetric(true)
{
    TEST_EXIT(phaseDV_)("no vector phase!\n");
    auxFeSpaces.insert(phaseDV_->getFeSpace());
};
void MatrixPhase_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void MatrixPhase_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};

/// Evaluation of \f$ \Lambda \cdot A \cdot \Lambda^t\f$ for A equal to the diagonal matrix.
void MatrixPhase_SOT::lDlt(const DimVec<WorldVector<double> >& Lambda, 
            const WorldVector<double>& vec,
            mtl::dense2D<double>& LALt,
            double factor) const
{
    const int dim = num_rows(LALt);
    
    for (int i = 0; i < dim; i++) {
        double val = 0.0;
        for (int k = 0; k < dimOfWorld; k++)
            val += vec[k] * Lambda[i][k] * Lambda[i][k];
        val *= factor;
        LALt[i][i] += val;
        for (int j = i + 1; j < dim; j++) {
            val = 0.0;
            for (int k = 0; k < dimOfWorld; k++)
                val += vec[k] * Lambda[i][k] * Lambda[j][k];
            val *= factor;
            LALt[i][j] += val;
            LALt[j][i] += val;
        }    
    }      
};
void MatrixPhase_SOT::getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const  
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++) {
        lDlt(Lambda, f(iq), LALt[iq], phase[iq] * fac);
    }
};
void MatrixPhase_SOT::eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor)
{   
    if (num_rows(D2UhAtQP) > 0) {
        double factor = opFactor * fac;
        for (int iq = 0; iq < nPoints; iq++) {
            WorldVector<double> D = f(iq);
            double resultQP = 0.0;
                    
            for (int i = 0; i < dimOfWorld; i++) {
                double resultI= 0.0;
                for (int j = 0; j < dimOfWorld; j++) {
                    resultI += D2UhAtQP[iq][j][i];
                }
                resultQP+= D[i] * resultI;
            }
            result[iq] += resultQP * factor * phase[iq];
        }
    }
};
void MatrixPhase_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result)
{
    const int nPoints = static_cast<int>(grdUhAtQP.size());
    for (int iq = 0; iq < nPoints; iq++) {
        WorldVector<double> multVec = f(iq);
        for (int i = 0; i < grdUhAtQP[iq].getSize(); ++i)
            multVec[i]*= grdUhAtQP[iq][i];
        result[iq] += multVec * (phase[iq] * fac);
    }
};
WorldVector<double> MatrixPhase_SOT::f(const int iq) const
{
    WorldVector<double> result;
    result[1-comp]=1.0;
    result[comp]=2.0;
    return result;
};


/* ----------------------------------------------------------- */


MatrixIJ_SOT::MatrixIJ_SOT(int row_, int col_, double fac_) :
    SecondOrderTerm(1), row(row_), col(col_), fac(fac_), symmetric(true)
{
};
void MatrixIJ_SOT::getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const  
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++) {
        lalt_kl(Lambda, row, col, LALt[iq], fac);
    }
};
void MatrixIJ_SOT::eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor)
{   
    double factor = opFactor * fac;
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            result[iq] += D2UhAtQP[iq][row][col] * factor;
        }
    }
};
void MatrixIJ_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result)
{
    const int nPoints = static_cast<int>(grdUhAtQP.size());
    for (int iq = 0; iq < nPoints; iq++) { 
        result[iq][row] += grdUhAtQP[iq][col] * fac;      
    }
};

/* ----------------------------------------------------------- */


MatrixIJPhase_SOT::MatrixIJPhase_SOT(DOFVectorBase<double>* phaseDV_, int row_, int col_, double fac_) :
    SecondOrderTerm(6), phaseDV(phaseDV_), row(row_), col(col_), fac(fac_), symmetric(true)
{
    TEST_EXIT(phaseDV_)("no vector phase!\n");
    auxFeSpaces.insert(phaseDV_->getFeSpace());
};
void MatrixIJPhase_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
};
void MatrixIJPhase_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
};
void MatrixIJPhase_SOT::getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const  
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++) {
        lalt_kl(Lambda, row, col, LALt[iq], phase[iq] * fac);
    }
};
void MatrixIJPhase_SOT::eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor)
{   
    double factor = opFactor * fac;
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            result[iq] += D2UhAtQP[iq][row][col] * factor;
        }
    }
};
void MatrixIJPhase_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result)
{
    const int nPoints = static_cast<int>(grdUhAtQP.size());
    for (int iq = 0; iq < nPoints; iq++) { 
        result[iq][row] += grdUhAtQP[iq][col] * fac;      
    }
};

/* ----------------------------------------------------------- */

MatrixIJViscosityPhase_SOT::MatrixIJViscosityPhase_SOT(DOFVectorBase<double>* phaseDV_, DOFVectorBase<double>* chDV_, int row_, int col_, double v1_, double v2_, double fac_) :
    SecondOrderTerm(6), phaseDV(phaseDV_), chDV(chDV_), row(row_), col(col_), v1(v1_), v2(v2_), fac(fac_), symmetric(true)
{
    TEST_EXIT(phaseDV_)("no vector phase!\n");
    auxFeSpaces.insert(phaseDV_->getFeSpace());
    auxFeSpaces.insert(chDV_->getFeSpace());
};
void MatrixIJViscosityPhase_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void MatrixIJViscosityPhase_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
void MatrixIJViscosityPhase_SOT::getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const  
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++) {
        lalt_kl(Lambda, row, col, LALt[iq], phase[iq] * f(iq) * fac);
    }
};
void MatrixIJViscosityPhase_SOT::eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor)
{   
    double factor = opFactor * fac;
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            result[iq] += D2UhAtQP[iq][row][col] * phase[iq] * f(iq) * factor;
        }
    }
};
void MatrixIJViscosityPhase_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result)
{
    const int nPoints = static_cast<int>(grdUhAtQP.size());
    for (int iq = 0; iq < nPoints; iq++) { 
        result[iq][row] += grdUhAtQP[iq][col] * phase[iq] * f(iq) * fac;      
    }
};
double MatrixIJViscosityPhase_SOT::f(const int iq) const
{
    double phase= std::max(-1.0,std::min(1.0,ch[iq]));
    double viscosity= v1*0.5*(phase+1.0)-v2*0.5*(phase-1.0);
    return viscosity;
};

/* ----------------------------------------------------------- */

MatrixIJViscosity_SOT::MatrixIJViscosity_SOT(DOFVectorBase<double>* chDV_, int row_, int col_, double v1_, double v2_, double fac_) :
    SecondOrderTerm(6), chDV(chDV_), row(row_), col(col_), v1(v1_), v2(v2_), fac(fac_), symmetric(true)
{
    TEST_EXIT(chDV_)("no vector phase!\n");
    auxFeSpaces.insert(chDV_->getFeSpace());
};
void MatrixIJViscosity_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void MatrixIJViscosity_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
void MatrixIJViscosity_SOT::getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const  
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++) {
        lalt_kl(Lambda, row, col, LALt[iq], f(iq) * fac);
    }
};
void MatrixIJViscosity_SOT::eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor)
{   
    double factor = opFactor * fac;
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            result[iq] += D2UhAtQP[iq][row][col] * f(iq) * factor;
        }
    }
};
void MatrixIJViscosity_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result)
{
    const int nPoints = static_cast<int>(grdUhAtQP.size());
    for (int iq = 0; iq < nPoints; iq++) { 
        result[iq][row] += grdUhAtQP[iq][col] * f(iq) * fac;      
    }
};
double MatrixIJViscosity_SOT::f(const int iq) const
{
    double phase= std::max(-1.0,std::min(1.0,ch[iq]));
    double viscosity= v1*0.5*(phase+1.0)-v2*0.5*(phase-1.0);
    return viscosity;
};

/* ----------------------------------------------------------- */


MatrixIJViscosityPhase2_SOT::MatrixIJViscosityPhase2_SOT(DOFVectorBase<double>* phaseDV_, DOFVectorBase<double>* chDV_, int row_, int col_, double v1_, double v2_, double fac_) :
    SecondOrderTerm(6), phaseDV(phaseDV_), chDV(chDV_), row(row_), col(col_), v1(v1_), v2(v2_), fac(fac_), symmetric(true)
{
    TEST_EXIT(phaseDV_)("no vector phase!\n");
    auxFeSpaces.insert(phaseDV_->getFeSpace());
    auxFeSpaces.insert(chDV_->getFeSpace());
};
void MatrixIJViscosityPhase2_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, elInfo, subAssembler, quad, phase);
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void MatrixIJViscosityPhase2_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature *quad) {
    getVectorAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, phase);
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
void MatrixIJViscosityPhase2_SOT::getLALt(const ElInfo *elInfo,
        std::vector<mtl::dense2D<double> > &LALt) const  
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++) {
        lalt_kl(Lambda, row, col, LALt[iq], phase[iq] * f(iq) * fac);
    }
};
void MatrixIJViscosityPhase2_SOT::eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor)
{   
    double factor = opFactor * fac;
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            result[iq] += D2UhAtQP[iq][row][col] * phase[iq] * f(iq) * factor;
        }
    }
};
void MatrixIJViscosityPhase2_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
        std::vector<WorldVector<double> > &result)
{
    const int nPoints = static_cast<int>(grdUhAtQP.size());
    for (int iq = 0; iq < nPoints; iq++) { 
        result[iq][row] += grdUhAtQP[iq][col] * phase[iq] * f(iq) * fac;      
    }
};
double MatrixIJViscosityPhase2_SOT::f(const int iq) const
{
    double phase= std::max(0.0,std::min(1.0,ch[iq]));
    double viscosity= v1*phase+v2*(1.0-phase);
    return viscosity;
};

/* ----------------------------------------------------------- */

WettingAngle_ZOT::WettingAngle_ZOT(DOFVectorBase<double> *vecDV_, DOFVectorBase<double> *gradDV_, double fac_)
	: ZeroOrderTerm(1), vecDV(vecDV_), gradDV(gradDV_), fac(fac_)
{
	TEST_EXIT(vecDV_)("No first vector!\n");
	auxFeSpaces.insert(vecDV_->getFeSpace());
	auxFeSpaces.insert(gradDV_->getFeSpace());
};
void WettingAngle_ZOT::initElement(const ElInfo* elInfo, 
					SubAssembler* subAssembler,
					Quadrature *quad)
{
// 	getVectorAtQPs(vecDV, elInfo, subAssembler, quad, vec);
        getGradientsAtQPs(vecDV, elInfo, subAssembler, quad, gradC);
	getGradientsAtQPs(gradDV, elInfo, subAssembler, quad, grad);
};
void WettingAngle_ZOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
					SubAssembler* subAssembler,
					Quadrature *quad)
{
// 	getVectorAtQPs(vecDV, smallElInfo, largeElInfo, subAssembler, quad, vec);
        getGradientsAtQPs(vecDV, smallElInfo, largeElInfo, subAssembler, quad, gradC);
	getGradientsAtQPs(gradDV, smallElInfo, largeElInfo, subAssembler, quad, grad);
};
void WettingAngle_ZOT::getC(const ElInfo *, int nPoints, ElementVector &C)
{ 	
	for (int iq = 0; iq < nPoints; iq++)
		C[iq] += f(iq)*fac;
};
void WettingAngle_ZOT::eval(int nPoints,
				const mtl::dense_vector<double>& uhAtQP,
				const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				mtl::dense_vector<double>& result,
				double opFactor)
{	
	for (int iq = 0; iq < nPoints; iq++) {		
		result[iq] += f(iq) * uhAtQP[iq] * opFactor * fac;
	}
};
double WettingAngle_ZOT::f(const int iq) const
{
//     double c= std::max(-1.0, std::min(1.0, vec[iq]) );
    return norm(gradC[iq])*norm(grad[iq]);
};

/* ----------------------------------------------------------- */

WettingAngleExpl_ZOT::WettingAngleExpl_ZOT(DOFVectorBase<double> *vecDV_, DOFVectorBase<double> *gradDV_, double fac_)
	: WettingAngle_ZOT(vecDV_,gradDV_,fac_)
{
  ERROR("Not implementet!\n");
};
double WettingAngleExpl_ZOT::f(const int iq) const
{
//     double c= std::max(-1.0, std::min(1.0, vec[iq]) );
    return 1.0*norm(grad[iq]);
};

/* ----------------------------------------------------------- */

WettingAngleImpl_ZOT::WettingAngleImpl_ZOT(DOFVectorBase<double> *vecDV_, DOFVectorBase<double> *gradDV_, double fac_)
	: WettingAngle_ZOT(vecDV_,gradDV_,fac_)
{
  ERROR("Not implementet!\n");
};
double WettingAngleImpl_ZOT::f(const int iq) const
{
//     double c= std::max(-1.0, std::min(1.0, vec[iq]) );
    return 2.0*norm(grad[iq]);
};


/* ----------------------------------------------------------- */

WettingAngleImpl_FOT::WettingAngleImpl_FOT(DOFVectorBase<double>* chDV_,DOFVectorBase<double>* phaseDV_,double fac_) :
        FirstOrderTerm(2), chDV(chDV_), phaseDV(phaseDV_), fac(fac_)
{
        TEST_EXIT(phaseDV_)("no vector phase!\n");
        auxFeSpaces.insert(chDV_->getFeSpace());
        auxFeSpaces.insert(phaseDV_->getFeSpace());
};
void WettingAngleImpl_FOT::initElement(const ElInfo* elInfo, 
                        SubAssembler* subAssembler,
                        Quadrature *quad) {
  getGradientsAtQPs(chDV, elInfo, subAssembler, quad, gradCh);
  getGradientsAtQPs(phaseDV, elInfo, subAssembler, quad, gradPhase);
};
void WettingAngleImpl_FOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
                        SubAssembler* subAssembler,
                        Quadrature *quad) {
  getGradientsAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, gradCh);
  getGradientsAtQPs(phaseDV, smallElInfo, largeElInfo, subAssembler, quad, gradPhase);
};
void WettingAngleImpl_FOT::getLb(const ElInfo *elInfo, 
            vector<mtl::dense_vector<double> >& result) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(result.size());
    
    for (int iq = 0; iq < nPoints; iq++) {
        lb(Lambda, f(iq), result[iq], fac);
    }
};
void WettingAngleImpl_FOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    double factor = fac * opFactor;
    for (int iq = 0; iq < nPoints; iq++) {
        double resultQP = grdUhAtQP[iq] * f(iq);
        result[iq] += factor * resultQP;
    }
};

WorldVector<double> WettingAngleImpl_FOT::f(const int iq) const
{
    WorldVector<double> result; result.set(0.0);
    double nrmGradCh = norm(gradCh[iq]);
    if (nrmGradCh > DBL_TOL) {
      result = (2.0 * norm(gradPhase[iq]) / nrmGradCh) * gradCh[iq];
    }
    return result;
};

/* ----------------------------------------------------------- */
/* ----------------------------------------------------------- */

WettingAngle_Sharp_ZOT::WettingAngle_Sharp_ZOT(DOFVectorBase<double> *vecDV_, double fac_)
  : ZeroOrderTerm(1), vecDV(vecDV_), fac(fac_)
{
  TEST_EXIT(vecDV_)("No first vector!\n");
  auxFeSpaces.insert(vecDV_->getFeSpace());
};
void WettingAngle_Sharp_ZOT::initElement(const ElInfo* elInfo, 
          SubAssembler* subAssembler,
          Quadrature *quad)
{
  getGradientsAtQPs(vecDV, elInfo, subAssembler, quad, grad);
};
void WettingAngle_Sharp_ZOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
          SubAssembler* subAssembler,
          Quadrature *quad)
{
  getGradientsAtQPs(vecDV, smallElInfo, largeElInfo, subAssembler, quad, grad);
};
void WettingAngle_Sharp_ZOT::getC(const ElInfo *, int nPoints, ElementVector &C)
{   
  for (int iq = 0; iq < nPoints; iq++)
    C[iq] += f(iq)*fac;
};
void WettingAngle_Sharp_ZOT::eval(int nPoints,
        const mtl::dense_vector<double>& uhAtQP,
        const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
        const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
        mtl::dense_vector<double>& result,
        double opFactor)
{ 
  for (int iq = 0; iq < nPoints; iq++) {    
    result[iq] += f(iq) * uhAtQP[iq] * opFactor * fac;
  }
};
double WettingAngle_Sharp_ZOT::f(const int iq) const
{
    double c= norm(grad[iq]);
    return c;
};

/* ----------------------------------------------------------- */

WettingAngleExpl_Sharp_ZOT::WettingAngleExpl_Sharp_ZOT(DOFVectorBase<double> *vecDV_, double fac_)
  : WettingAngle_Sharp_ZOT(vecDV_,fac_)
{
  ERROR("Not implementet!\n");
};
double WettingAngleExpl_Sharp_ZOT::f(const int iq) const
{
//     double c= std::max(-1.0, std::min(1.0, vec[iq]) );
    return 1.0;
};

/* ----------------------------------------------------------- */

WettingAngleImpl_Sharp_ZOT::WettingAngleImpl_Sharp_ZOT(DOFVectorBase<double> *vecDV_, double fac_)
  : WettingAngle_Sharp_ZOT(vecDV_,fac_)
{
  ERROR("Not implementet!\n");
};
double WettingAngleImpl_Sharp_ZOT::f(const int iq) const
{
//     double c= std::max(-1.0, std::min(1.0, vec[iq]) );
    return 2.0;
};


/* ----------------------------------------------------------- */

ViscosityPhase_SOT::ViscosityPhase_SOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *chDV_, double v1_, double v2_, double factor_)
    : Phase_SOT(phaseDV_,factor_), chDV(chDV_), v1(v1_), v2(v2_)
{
    setSymmetric(true);
    
    TEST_EXIT(chDV)("No vector!\n");
    auxFeSpaces.insert(chDV->getFeSpace());
};
void ViscosityPhase_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
	Phase_SOT::initElement(elInfo, subAssembler, quad);
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void ViscosityPhase_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
	Phase_SOT::initElement(smallElInfo, largeElInfo, subAssembler, quad);
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
double ViscosityPhase_SOT::f(const int iq) const
{
    double phase= std::max(-1.0,std::min(1.0,ch[iq]));
    double viscosity= v1*0.5*(phase+1.0)-v2*0.5*(phase-1.0);
    return viscosity;
};

/* ----------------------------------------------------------- */

Viscosity_SOT::Viscosity_SOT(DOFVectorBase<double> *chDV_, double v1_, double v2_, double factor_)
    : SecondOrderTerm(1), chDV(chDV_), v1(v1_), v2(v2_), factor(factor_)
{
    setSymmetric(true);
    
    TEST_EXIT(chDV)("No vector!\n");
    auxFeSpaces.insert(chDV->getFeSpace());
};
void Viscosity_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void Viscosity_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
void Viscosity_SOT::getLALt(const ElInfo *elInfo,
            std::vector<mtl::dense2D<double> > &LALt) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());
    
    for (int iq = 0; iq < nPoints; iq++)
        l1lt(Lambda, LALt[iq], f(iq) * factor);
};
void Viscosity_SOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            double feval = f(iq) * opFactor * factor;
            double resultQP = 0.0;
            for (int i = 0; i < dimOfWorld; i++)
                resultQP += D2UhAtQP[iq][i][i]; 
            result[iq] += feval * resultQP;
        }
    }
};
void Viscosity_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
            std::vector<WorldVector<double> > &result)
{
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
        double fac = f(iq) * factor;
        axpy(fac, grdUhAtQP[iq], result[iq]);
    }
};
double Viscosity_SOT::f(const int iq) const
{
    double phase= std::max(-1.0,std::min(1.0,ch[iq]));
    double viscosity= v1*0.5*(phase+1.0)-v2*0.5*(phase-1.0);
    return viscosity;
};

/* ----------------------------------------------------------- */

/* ----------------------------------------------------------- */

ViscosityPhase2_SOT::ViscosityPhase2_SOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *chDV_, double v1_, double v2_, double factor_)
    : Phase_SOT(phaseDV_,factor_), chDV(chDV_), v1(v1_), v2(v2_)
{
    setSymmetric(true);
    
    TEST_EXIT(chDV)("No vector!\n");
    auxFeSpaces.insert(chDV->getFeSpace());
};
void ViscosityPhase2_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
  Phase_SOT::initElement(elInfo, subAssembler, quad);
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void ViscosityPhase2_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
  Phase_SOT::initElement(smallElInfo, largeElInfo, subAssembler, quad);
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
double ViscosityPhase2_SOT::f(const int iq) const
{
    double phase= std::max(0.0,std::min(1.0,ch[iq]));
    double viscosity= v1*phase+v2*(1.0-phase);
    return viscosity;
};

/* ----------------------------------------------------------- */

Viscosity2_SOT::Viscosity2_SOT(DOFVectorBase<double> *chDV_, double v1_, double v2_, double factor_)
    : SecondOrderTerm(1), chDV(chDV_), v1(v1_), v2(v2_), factor(factor_)
{
    setSymmetric(true);
    
    TEST_EXIT(chDV)("No vector!\n");
    auxFeSpaces.insert(chDV->getFeSpace());
};
void Viscosity2_SOT::initElement(const ElInfo* elInfo, 
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, elInfo, subAssembler, quad, ch);
};
void Viscosity2_SOT::initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
            SubAssembler* subAssembler,
            Quadrature* quad) 
{
    getVectorAtQPs(chDV, smallElInfo, largeElInfo, subAssembler, quad, ch);
};
void Viscosity2_SOT::getLALt(const ElInfo *elInfo,
            std::vector<mtl::dense2D<double> > &LALt) const
{
    const DimVec<WorldVector<double> > &Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());
    
    for (int iq = 0; iq < nPoints; iq++)
        l1lt(Lambda, LALt[iq], f(iq) * factor);
};
void Viscosity2_SOT::eval(int nPoints,
            const mtl::dense_vector<double>& uhAtQP,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor)
{   
    if (num_rows(D2UhAtQP) > 0) {
        for (int iq = 0; iq < nPoints; iq++) {
            double feval = f(iq) * opFactor * factor;
            double resultQP = 0.0;
            for (int i = 0; i < dimOfWorld; i++)
                resultQP += D2UhAtQP[iq][i][i]; 
            result[iq] += feval * resultQP;
        }
    }
};
void Viscosity2_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
            std::vector<WorldVector<double> > &result)
{
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
        double fac = f(iq) * factor;
        axpy(fac, grdUhAtQP[iq], result[iq]);
    }
};
double Viscosity2_SOT::f(const int iq) const
{
    double phase= std::max(0.0,std::min(1.0,ch[iq]));
    double viscosity= v1*phase+v2*(1.0-phase);
    return viscosity;
};
