#include "LinearElasticity.h"

using namespace std;
using namespace AMDiS;

LinearElasticity::LinearElasticity(const std::string &name_) :
  super(name_),
  mu(1.0),
  lambda(1.0),
  rho(1.0),
  displacement(NULL)
{  
  Parameters::get(name + "->mu", mu);
  Parameters::get(name + "->lambda", lambda);
  Parameters::get(name + "->rho", rho);
  
  force.set(0.0);
  Parameters::get(name + "->force", force);
}


LinearElasticity::~LinearElasticity() 
{ FUNCNAME("LinearElasticity::~LinearElasticity()");

  if (displacement != NULL)
    delete displacement;
  delete fileWriter;
}


void LinearElasticity::initData()
{ FUNCNAME("LinearElasticity::initTimeInterface()");

  if (displacement == NULL)
    displacement = new DOFVector<WorldVector<double> >(getFeSpace(0), "displacement");
  fileWriter = new FileVectorWriter(name + "->displacement->output", getFeSpace()->getMesh(), displacement);

  super::initData();
}


void LinearElasticity::solveInitialProblem(AdaptInfo *adaptInfo) 
{ FUNCNAME("LinearElasticity::solveInitialProblem()");

  for (size_t i = 0; i < 2*dow-1; i++)
    prob->getSolution()->getDOFVector(i)->set(0.0);
}


void LinearElasticity::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("LinearElasticity::transferInitialSolution()");

  if (dow == 1)
    transformDOF(prob->getSolution()->getDOFVector(0), displacement, new AMDiS::Vec1WorldVec<double>);
  else if (dow == 2)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), displacement, new AMDiS::Vec2WorldVec<double>);
  else if (dow == 3)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), prob->getSolution()->getDOFVector(2), displacement, new AMDiS::Vec3WorldVec<double>);

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);
  
  // initial parameters for detecting mesh changes
  oldMeshChangeIdx= getMesh()->getChangeIndex();
}


void LinearElasticity::fillOperators()
{ FUNCNAME("LinearElasticity::fillOperators()");
  MSG("LinearElasticity::fillOperators()\n");

  // fill operators for prob
  for (size_t i = 0; i < dow; ++i) {
    /// dt(u) - v = 0
    /// ____________________________________________________________________________
    
    if (abs(rho) > DBL_TOL) {
      // < (1/tau)*u_i , psi >
      Operator *opTimeU = new Operator(getFeSpace(i+dow), getFeSpace(i));
      opTimeU->addTerm(new Simple_ZOT);
      prob->addMatrixOperator(*opTimeU, i+dow, i, getInvTau(), getInvTau());
      // < (1/tau)*u_i^old , psi >
      Operator *opTimeUOld = new Operator(getFeSpace(i+dow));
      opTimeUOld->addTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(i)));
      prob->addVectorOperator(*opTimeUOld, i+dow, getInvTau(), getInvTau());
    }
    // < v , psi >
    Operator *opV = new Operator(getFeSpace(i+dow), getFeSpace(i+dow));
    opV->addTerm(new Simple_ZOT(abs(rho) > DBL_TOL ? -1.0/rho : -1.0));
    prob->addMatrixOperator(*opV, i+dow, i+dow);
    
    /// dt(v) - mu*div(grad(u)) - (mu+lambda)*grad(div(u)) = F
    /// ____________________________________________________________________________
 
    
    if (abs(rho) > DBL_TOL) {
      // < (1/tau)*v_i , psi >
      Operator *opTimeV = new Operator(getFeSpace(i), getFeSpace(i+dow));
      opTimeV->addTerm(new Simple_ZOT);
      prob->addMatrixOperator(*opTimeV, i, i+dow, getInvTau(), getInvTau());
      // < (1/tau)*v_i^old , psi >
      Operator *opTimeVOld = new Operator(getFeSpace(i));
      opTimeVOld->addTerm(new VecAtQP_ZOT(prob->getSolution()->getDOFVector(i+dow)));
      prob->addVectorOperator(*opTimeVOld, i, getInvTau(), getInvTau());
    }
    
    // < grad(u_i) , grad(psi) >
    Operator *opDivGradU = new Operator(getFeSpace(i), getFeSpace(i));
    opDivGradU->addTerm(new Simple_SOT(mu));
    prob->addMatrixOperator(*opDivGradU, i, i);

    // < d_j(u_j) , d_i(psi) >
    for (size_t j = 0; j < dow; j++) {
      Operator *opGradDivU = new Operator(getFeSpace(i), getFeSpace(j));
      opGradDivU->addTerm(new FactorIJ_SOT(i, j, mu+lambda));
      prob->addMatrixOperator(*opGradDivU, i, j);
    }
  
    // external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      opForce->addTerm(new Simple_ZOT(force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }
}


void LinearElasticity::fillBoundaryConditions()
{ FUNCNAME("LinearElasticity::fillBoundaryConditions()");

//   ERROR("You have to implement some boundary conditions!\n");
  AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
  for (size_t i = 0; i < 2*dow-1; i++) {
    prob->addDirichletBC(1, i, i, zero);
  }
}


// void LinearElasticity::initTimestep(AdaptInfo *adaptInfo)
// { FUNCNAME("LinearElasticity::initTimestep()");
// 
// }


void LinearElasticity::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("LinearElasticity::closeTimestep()");

  if (dow == 1)
    transformDOF(prob->getSolution()->getDOFVector(0), displacement, new AMDiS::Vec1WorldVec<double>);
  else if (dow == 2)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), displacement, new AMDiS::Vec2WorldVec<double>);
  else if (dow == 3)
    transformDOF(prob->getSolution()->getDOFVector(0), prob->getSolution()->getDOFVector(1), prob->getSolution()->getDOFVector(2), displacement, new AMDiS::Vec3WorldVec<double>);

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);
}
