//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "time/ExtendedRosenbrockAdaptInstationary.h"
#include "AdaptInfo.h"
#include "ProblemTimeInterface.h"
#include "VectorOperations.h"

namespace AMDiS {

  template<typename ProblemStationary>
  ExtendedRosenbrockAdaptInstationary<ProblemStationary>::ExtendedRosenbrockAdaptInstationary(std::string name,
							   ProblemStationary *problemStat,
							   AdaptInfo *info,
							   ProblemTimeInterface *problemInstat,
							   AdaptInfo *initialInfo,
							   time_t initialTimestamp)
    : AdaptInstationary(name, problemStat, info, problemInstat, initialInfo, initialTimestamp),
      rosenbrockStat(problemStat),
      firstTimestep(true),
      lastTimestepRejected(false),
      succRejection(false),
      fixFirstTimesteps(0),
      tau(1.0),
      tauGamma(1.0),
      minusTauGamma(-1.0),
      invTauGamma(1.0),
      minusInvTauGamma(-1.0),
      dbgTimestepStudy(false)
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::ExtendedRosenbrockAdaptInstationary()");
    initConstructor(problemStat);
  }


  template<typename ProblemStationary>
  ExtendedRosenbrockAdaptInstationary<ProblemStationary>::ExtendedRosenbrockAdaptInstationary(std::string name,
							   ProblemStationary &problemStat,
							   AdaptInfo &info,
							   ProblemTimeInterface &problemInstat,
							   AdaptInfo &initialInfo,
							   time_t initialTimestamp)
    : AdaptInstationary(name, problemStat, info, problemInstat, initialInfo, initialTimestamp),
      rosenbrockStat(&problemStat),
      firstTimestep(true),
      lastTimestepRejected(false),
      succRejection(false),
      fixFirstTimesteps(0),
      tau(1.0),
      tauGamma(1.0),
      minusTauGamma(-1.0),
      invTauGamma(1.0),
      minusInvTauGamma(-1.0),
      dbgTimestepStudy(false)
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::ExtendedRosenbrockAdaptInstationary()");
    initConstructor(&problemStat);
  }


  template<typename ProblemStationary>
  void ExtendedRosenbrockAdaptInstationary<ProblemStationary>::initConstructor(ProblemStationary *problemStat)
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::initConstructor()");

    std::string str("");
    std::string initFileStr(name + "->rosenbrock method");
    Parameters::get(initFileStr, str);
    RosenbrockMethodCreator *creator = 
      dynamic_cast<RosenbrockMethodCreator*>(CreatorMap<RosenbrockMethod>::getCreator(str, initFileStr));
    rosenbrockMethod = creator->create();
    
    TEST_EXIT_DBG(rosenbrockMethod)("This should not happen!\n");
    
    Parameters::get(name + "->fix first timesteps", fixFirstTimesteps);
    rosenbrockStat->setRosenbrockMethod(rosenbrockMethod);
    
    adaptInfo->setRosenbrockMode(true);
    
    rosenbrockStat->setTauGamma(&tauGamma, &minusTauGamma,
			     &invTauGamma, &minusInvTauGamma);
    rosenbrockStat->setTau(&tau);
    
    Parameters::get(name + "->rosenbrock->timestep study", dbgTimestepStudy);
    Parameters::get(name + "->rosenbrock->timestep study steps", dbgTimesteps);

    Parameters::get(name + "->rosenbrock->error weights", errorEstWeights);
    if (errorEstWeights.size() == 0) {
      WARNING("No weights for time-error estimation given.\n");
    }

    errorEstWeights.resize(adaptInfo->getSize(), 0.0);
    double SumErrorEstWeights = vector_operations::sum(errorEstWeights);
    if (SumErrorEstWeights<DBL_TOL) {
      WARNING("Using equal weights for all components!\n");
      for (size_t i = 0; i < errorEstWeights.size(); i++)
	errorEstWeights[i] = 1.0/static_cast<double>(adaptInfo->getSize());
    } else {
      for (size_t i = 0; i < errorEstWeights.size(); i++)
	errorEstWeights[i] *= 1.0/SumErrorEstWeights;
    }
  }


  template<typename ProblemStationary>
  void ExtendedRosenbrockAdaptInstationary<ProblemStationary>::reset()
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::reset()");

    firstTimestep = true;
    lastTimestepRejected = false;
    succRejection = false;
    fixFirstTimesteps = 0;
    tau = 1.0;
    tauGamma = 1.0;    
    minusTauGamma = -1.0;
    invTauGamma = 1.0;
    minusInvTauGamma = -1.0;
  }


  template<typename ProblemStationary>
  void ExtendedRosenbrockAdaptInstationary<ProblemStationary>::oneTimestep()
  {
    FUNCNAME("ExtendedRosenbrockAdaptInstationary::oneTimestep()");
    
    // estimate before first adaption
    if (adaptInfo->getTime() <= adaptInfo->getStartTime())
      problemIteration->oneIteration(adaptInfo, ESTIMATE);
    
    bool rejected = false;
    double timeTol = adaptInfo->getTimeTolerance(0);

    int studyTimestep = -1;

    do {
      if (dbgTimestepStudy) {
	if (studyTimestep >= 0)
	  adaptInfo->setTime(adaptInfo->getTime() - dbgTimesteps[studyTimestep]);

	studyTimestep++;
	TEST_EXIT(studyTimestep < dbgTimesteps.size())("Should not happen!\n");
	
	adaptInfo->setTimestep(dbgTimesteps[studyTimestep]);
      }

      // increment time
      adaptInfo->setTime(adaptInfo->getTime() + adaptInfo->getTimestep());
      problemTime->setTime(adaptInfo);
      tau = adaptInfo->getTimestep();

      tauGamma = tau * rosenbrockMethod->getGamma();
      minusTauGamma = -tauGamma;
      invTauGamma = 1.0 / (tau * rosenbrockMethod->getGamma());
      minusInvTauGamma = -1.0;
      
      INFO(info, 6)("time = %e   timestep = %e\n",
		    adaptInfo->getTime(), adaptInfo->getTimestep());
      
      adaptInfo->setSpaceIteration(0);
      
      // do the iteration
      problemIteration->beginIteration(adaptInfo);
      problemIteration->oneIteration(adaptInfo, FULL_ITERATION);
      problemIteration->endIteration(adaptInfo);

      double errorEst = 0.0;
      for (size_t i = 0; i < adaptInfo->getSize(); i++) {
	errorEst += errorEstWeights[i] * adaptInfo->getTimeEstSum(i);
      }
      
      double newTimestep = 0.0;
      double order = rosenbrockMethod->getOrder();

      if (errorEst < timeTol || fixFirstTimesteps > 0 || firstTimestep) {
	double fac = 1.0;
	
	if (fixFirstTimesteps > 0) {
	  newTimestep = adaptInfo->getTimestep();
	} else {
	  if (firstTimestep || succRejection) {
	    fac = pow((timeTol / errorEst), 1.0 / order);
	  } else {
	    fac = adaptInfo->getTimestep() / tauAcc * 
	      pow((timeTol * estAcc / (errorEst * errorEst)), 1.0 / order);
	  }
	  fac = std::min(fac, 3.0);
	  newTimestep = fac * adaptInfo->getTimestep();
	  newTimestep *= 0.95;
	}

	tauAcc = adaptInfo->getTimestep();
	estAcc = errorEst;
	
	lastTimestepRejected = false;
	succRejection = false;
      } else {
	if (lastTimestepRejected) {
	  succRejection = true;  
	  
	  double reducedP = 
	    log(errorEst / estRej) / log(adaptInfo->getTimestep() / tauRej);
	  
	  if (reducedP < order && reducedP > 0.0)
	    order = reducedP;
	} 
	
	newTimestep = pow((timeTol / errorEst), 1.0 / order) * adaptInfo->getTimestep();
	newTimestep *= 0.95;

	tauRej = adaptInfo->getTimestep();
	estRej = errorEst;
	
	lastTimestepRejected = true;
      }
      
      
      if (errorEst < timeTol || fixFirstTimesteps 
	  || !(adaptInfo->getTimestep()>adaptInfo->getMinTimestep()) ) {
	INFO(info, 6)("Accepted timestep at time = %e   with timestep = %e\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());

	adaptInfo->setTimestep(newTimestep);

	rejected = false;

	for (int i = 0; i < adaptInfo->getSize(); i++)
	  INFO(info, 6)("time estimate for component %d = %e\n", 
			i, adaptInfo->getTimeEstSum(i));

        if (errorEst > timeTol)
	  MSG("Accepted timestep but tolerance not reached \n");

      } else {
	for (int i = 0; i < adaptInfo->getSize(); i++)
	  INFO(info, 6)("time estimate for component %d = %e\n", 
			i, adaptInfo->getTimeEstSum(i));

	INFO(info, 6)("Rejected timestep at time = %e   with timestep = %e\n",
		      adaptInfo->getTime()  - adaptInfo->getTimestep(),
		      adaptInfo->getTimestep());
	
	adaptInfo->setTime(adaptInfo->getTime() - adaptInfo->getTimestep());
	adaptInfo->setTimestep(newTimestep);

	rejected = true;
      }

      if (firstTimestep)
	firstTimestep = false;

      if (fixFirstTimesteps > 0)
	fixFirstTimesteps--;     
    } while (rejected || 
	     (dbgTimestepStudy && (studyTimestep + 1 < dbgTimesteps.size())));

    rosenbrockStat->acceptTimestep();

    adaptInfo->setLastProcessedTimestep(adaptInfo->getTimestep());
    adaptInfo->incTimestepNumber();
  }

}
